// ----------------------------------------------------------------------------
// Copyright (c) 2014, Nicolas P. Rougier. All Rights Reserved.
// Distributed under the (new) BSD License.
// ----------------------------------------------------------------------------

#include "arrowheads/util.glsl"

#include "arrowheads/curved.glsl"
#include "arrowheads/stealth.glsl"
#include "arrowheads/triangle.glsl"
#include "arrowheads/angle.glsl"
#include "arrowheads/inhibitor.glsl"
